/*
 * Copyright 2011 Google Inc. All Rights Reserved.
 * Author: yugui@google.com (Yugui Sonoda)
 * */
#ifndef RUBY_NACL_RESOURCE_H
#define RUBY_NACL_RESOURCE_H
int getrusage(int who, struct rusage *usage);
#endif
