/* -*-c-*- *********************************************************/
/*******************************************************************/
/*******************************************************************/
/**
  This file is for threaded code.

  ----
  This file is auto generated by insns2vm.rb
  DO NOT TOUCH!

  If you want to fix something, you must edit 'template/vmtc.inc.tmpl'
  or insns2vm.rb
 */

static const void *const insns_address_table[] = {
  LABEL_PTR(nop),
  LABEL_PTR(getlocal),
  LABEL_PTR(setlocal),
  LABEL_PTR(getspecial),
  LABEL_PTR(setspecial),
  LABEL_PTR(getinstancevariable),
  LABEL_PTR(setinstancevariable),
  LABEL_PTR(getclassvariable),
  LABEL_PTR(setclassvariable),
  LABEL_PTR(getconstant),
  LABEL_PTR(setconstant),
  LABEL_PTR(getglobal),
  LABEL_PTR(setglobal),
  LABEL_PTR(putnil),
  LABEL_PTR(putself),
  LABEL_PTR(putobject),
  LABEL_PTR(putspecialobject),
  LABEL_PTR(putiseq),
  LABEL_PTR(putstring),
  LABEL_PTR(concatstrings),
  LABEL_PTR(tostring),
  LABEL_PTR(freezestring),
  LABEL_PTR(toregexp),
  LABEL_PTR(newarray),
  LABEL_PTR(duparray),
  LABEL_PTR(expandarray),
  LABEL_PTR(concatarray),
  LABEL_PTR(splatarray),
  LABEL_PTR(newhash),
  LABEL_PTR(newrange),
  LABEL_PTR(pop),
  LABEL_PTR(dup),
  LABEL_PTR(dupn),
  LABEL_PTR(swap),
  LABEL_PTR(reverse),
  LABEL_PTR(reput),
  LABEL_PTR(topn),
  LABEL_PTR(setn),
  LABEL_PTR(adjuststack),
  LABEL_PTR(defined),
  LABEL_PTR(checkmatch),
  LABEL_PTR(checkkeyword),
  LABEL_PTR(trace),
  LABEL_PTR(defineclass),
  LABEL_PTR(send),
  LABEL_PTR(opt_str_freeze),
  LABEL_PTR(opt_newarray_max),
  LABEL_PTR(opt_newarray_min),
  LABEL_PTR(opt_send_without_block),
  LABEL_PTR(invokesuper),
  LABEL_PTR(invokeblock),
  LABEL_PTR(leave),
  LABEL_PTR(throw),
  LABEL_PTR(jump),
  LABEL_PTR(branchif),
  LABEL_PTR(branchunless),
  LABEL_PTR(branchnil),
  LABEL_PTR(getinlinecache),
  LABEL_PTR(setinlinecache),
  LABEL_PTR(once),
  LABEL_PTR(opt_case_dispatch),
  LABEL_PTR(opt_plus),
  LABEL_PTR(opt_minus),
  LABEL_PTR(opt_mult),
  LABEL_PTR(opt_div),
  LABEL_PTR(opt_mod),
  LABEL_PTR(opt_eq),
  LABEL_PTR(opt_neq),
  LABEL_PTR(opt_lt),
  LABEL_PTR(opt_le),
  LABEL_PTR(opt_gt),
  LABEL_PTR(opt_ge),
  LABEL_PTR(opt_ltlt),
  LABEL_PTR(opt_aref),
  LABEL_PTR(opt_aset),
  LABEL_PTR(opt_aset_with),
  LABEL_PTR(opt_aref_with),
  LABEL_PTR(opt_length),
  LABEL_PTR(opt_size),
  LABEL_PTR(opt_empty_p),
  LABEL_PTR(opt_succ),
  LABEL_PTR(opt_not),
  LABEL_PTR(opt_regexpmatch1),
  LABEL_PTR(opt_regexpmatch2),
  LABEL_PTR(opt_call_c_function),
  LABEL_PTR(bitblt),
  LABEL_PTR(answer),
  LABEL_PTR(getlocal_OP__WC__0),
  LABEL_PTR(getlocal_OP__WC__1),
  LABEL_PTR(setlocal_OP__WC__0),
  LABEL_PTR(setlocal_OP__WC__1),
  LABEL_PTR(putobject_OP_INT2FIX_O_0_C_),
  LABEL_PTR(putobject_OP_INT2FIX_O_1_C_),
};

