<% releases.each do |release| %>
## [<%= release.name %>](https://github.com/<%= repository %>/tree/<%= release.tag_name %>)

_<sup>released on <%= release.published_at %></sup>_

<%= release.body.gsub(/\r\n/, "\n") %>

<% end %>
